"""
It's common practice in Linux distributions to debundle vendored
Python packages and modules for many reasons. Usually it's done
by direct patching of imports of vendored packages. It's can be
very time-consuming to rediff such patches on new upstream release.

This module transparently proxies imports of bundled top level
module/package to a normal one (any imported package with that name)
according to PEP302 rules. Thereby, it eliminates the need for source
code patching.

How to use:
- don't modify any source files to fix imports
- remove top level directory for vendored packages
- put this file as __init__.py into that newly created root-directory
  and configure replaceable variables


For example, six and requests are vendored into path mypackage/vendored.
Then changes for RPM spec file would look like:
- filter no longer provided self-dependencies
  %filter_from_requires /python3(mypackage\.vendored\..*)/d

- add new dependencies on debundled packages
  %py3_requires six
  %py3_requires requests

- bundle this module :)
  VENDORED_PATH='mypackage/vendored'
  UNVENDORED_PATH="$VENDORED_PATH/__init__.py"
  rm -r "$VENDORED_PATH"
  mkdir "$VENDORED_PATH"
  cp "%SOURCE1" "$UNVENDORED_PATH"
  sed -i \
        -e 's/"botocore.vendored"/"mypackage.vendored"/' \
        -e 's/{"requests.packages"}/None/' \
        "$UNVENDORED_PATH"

Hooray, we are done.

Checkout for a newer version:
https://git.altlinux.org/people/slev/public/debundler.git
"""
from importlib.abc import Loader
import sys
import importlib
import importlib.util


def parse_name(fullname, roots):
    """Search package across roots"""
    if not fullname.startswith(tuple(roots)):
        return None, None

    for root in roots:
        pkg_root, _, pkg = fullname.partition(root)
        # start with root
        if not pkg_root:
            return root, pkg

    return None, None


class ProxyLoader(importlib.abc.Loader):
    def __init__(self, *args, **kwargs):
        self.root = kwargs.pop("root")
        self.pkg = kwargs.pop("pkg")
        self.target_spec = kwargs.pop("target_spec")
        self.fullname = self.root + self.pkg
        super().__init__(*args, **kwargs)

    def create_module(self, spec):
        if spec.name != self.fullname:
            return None
        mod = importlib.import_module(self.pkg)
        return mod

    def exec_module(self, module):
        pass

    def get_resource_reader(self, fullname):
        if fullname != self.fullname:
            raise ValueError
        return self.target_spec.loader.get_resource_reader(self.pkg)


class UnvendoredImporter:
    def __init__(self, root, fake_packages=None):
        """
        root defines a root of bundled packages. For example,
        if some project bundles other packages into 'package/vendored/'
        then the root would be 'package.vendored'.

        fake_packages defines modules that behave like package,
        for example, requests.packages dirty hacks an import system
        (import packages into module and add them into its namespace).
        """
        self.roots = {root if root.endswith(".") else root + "."}
        self.fake_packages = set()
        if fake_packages is not None:
            self.fake_packages.update(set(fake_packages))

    def install(self):
        for importer in sys.meta_path:
            if importer.__class__.__name__ == self.__class__.__name__:
                importer.add(self)
                break
        else:
            sys.meta_path.insert(0, self)

    def add(self, other):
        self.roots.update(other.roots)
        self.fake_packages.update(other.fake_packages)

    def find_spec(self, fullname, path, target=None):
        if path is None:
            return None

        root, pkg = parse_name(fullname, self.roots)
        if root is None:
            return None

        target_spec = importlib.util.find_spec(pkg)
        if target_spec is None:
            return None

        return importlib.util.spec_from_loader(
            fullname,
            ProxyLoader(root=root, pkg=pkg, target_spec=target_spec),
            is_package=(
                target_spec.submodule_search_locations is not None
                or pkg in self.fake_packages
            ),
        )


UnvendoredImporter(
    "botocore.vendored",
    fake_packages={"requests.packages"},
).install()
